{-# LANGUAGE LambdaCase #-}
-- | Little module to parse (a subset of) ASCIIMath
module Math where

import Data.Text (Text, pack, singleton)
import Data.Text.Lazy (toStrict)
import Data.These
import Data.Char (isAlphaNum)
import Data.Functor ((<&>))
import Data.Void (Void)
import Control.Monad (replicateM)
import Text.Pandoc.Definition(Inline(..), Pandoc, Format(..), MathType(..))
import Text.Pandoc.Walk (walk, walkM)
import Text.Megaparsec
import Text.Megaparsec.Char
import qualified Text.Megaparsec.Char.Lexer as L

import Lucid.Base
import Lucid.Html5


-- Rendering

-- Some MathML core elements

-- | Element used to enclose math block.
math_ :: Term a r => a -> r
math_ = term "math"

-- | Element used to render arbitrary text.
mtext_ :: Functor m => HtmlT m a -> HtmlT m a
mtext_ = makeElement "mtext"

-- | Element used to represent a symbolic name or arbitrary text that should be rendered as an identifier.
mi_ :: Functor m => HtmlT m a -> HtmlT m a
mi_ = makeElement "mi"

-- | Element used to represent a numeric literal.
mn_ :: Functor m => HtmlT m a -> HtmlT m a
mn_ = makeElement "mn"

-- | Element used to represent an operator.
mo_ :: Functor m => HtmlT m a -> HtmlT m a
mo_ = makeElement "mo"

-- | Element used to represent a blank space.
mspace_ :: Applicative m => HtmlT m ()
mspace_ = makeElementNoEnd "mspace"

-- | Element used to represent a string literal.
ms_ :: Functor m => HtmlT m a -> HtmlT m a
ms_ = makeElement "ms"

mtable_ :: Functor m => HtmlT m a -> HtmlT m a
mtable_ = makeElement "mtable"

mtr_ :: Functor m => HtmlT m a -> HtmlT m a
mtr_ = makeElement "mtr"

mtd_ :: Functor m => HtmlT m a -> HtmlT m a
mtd_ = makeElement "mtd"

-- | Element used to group together any number of sub-expressions.
mrow_ :: Functor m => HtmlT m a -> HtmlT m a
mrow_ = makeElement "mrow"

-- | Element used for fractions. Can also be used to mark fraction-like objects.
mfrac_ :: Term a r => a -> r
mfrac_ = term "mfrac"

-- | Element used to construct an expression with a root symbol.
msqrt_ :: Functor m => HtmlT m a -> HtmlT m a
msqrt_ = makeElement "msqrt"

-- | Element used to construct an expression with a root and an index.
--   It has exactly two children: it is a root of index the second child, applied to the first child.
mroot_ :: Functor m => HtmlT m a -> HtmlT m a
mroot_ = makeElement "mroot"

mpadded_ :: Functor m => HtmlT m a -> HtmlT m a
mpadded_ = makeElement "mpadded"

merror_ :: Functor m => HtmlT m a -> HtmlT m a
merror_ = makeElement "merror"

msup_ :: Functor m => HtmlT m a -> HtmlT m a
msup_ = makeElement "msup"

msub_ :: Functor m => HtmlT m a -> HtmlT m a
msub_ = makeElement "msub"

msubsup_ :: Functor m => HtmlT m a -> HtmlT m a
msubsup_ = makeElement "msub"

munderover_ :: Term a r => a -> r
munderover_ = term "munderover"

munder_ :: Term a r => a -> r
munder_ = term "munder"

mover_ :: Term a r => a -> r
mover_ = term "mover"

-- | Either @infix@, @prefix@ or @postfix@, property of an embellished operator.
form_ :: Text -> Attribute
form_ = makeAttribute "form"

lspace_ :: Text -> Attribute
lspace_ = makeAttribute "lspace"

rspace_ :: Text -> Attribute
rspace_ = makeAttribute "rspace"

display_ :: Text -> Attribute
display_ = makeAttribute "display"

------------

-- | AST for math expression, made to be close to MathML.
data Exp
  = MError Text
  | MText Text
  | MFrac [Attribute] Exp Exp
  | MIdent Text
  | MMultiScript
  | MRow [Exp]
  | MNum Text
  | MOp Text -- TODO (operator attributes)
  | MUnderOver [Attribute] (These Exp Exp) Exp
  | MSubSup (These Exp Exp) Exp
  | MRoot (Maybe Exp) Exp
  | MTable [[Exp]]
  | Empty
  deriving (Show, Eq)

testExp :: Exp
testExp = MRoot Nothing (MRow [MIdent "x", MOp "+", MNum "1"])

instance ToHtml Exp where

  toHtml (MError e) = merror_ (toHtml e)
  toHtml (MText t) = mtext_ (toHtml t)
  toHtml (MIdent i) = mi_ (toHtml i)
  toHtml (MFrac attrs a b) = mfrac_ attrs (toHtml a <> toHtml b)
  toHtml (MNum t) = mn_ (toHtml t)
  toHtml (MRow r) = mrow_ (foldMap toHtml r)
  toHtml (MOp t) = mo_ (toHtml t)
  toHtml Empty = mempty
  toHtml (MUnderOver attrs (This a)    t) = munder_ attrs (toHtml t <> toHtml a)
  toHtml (MUnderOver attrs (That b)    t) = mover_ attrs (toHtml t             <> toHtml b)
  toHtml (MUnderOver attrs (These a b) t) = munderover_ attrs (toHtml t <> toHtml a <> toHtml b)
  toHtml (MRoot Nothing t)  = msqrt_ (toHtml t)
  toHtml (MRoot (Just r) t) = mroot_ (toHtml t <> toHtml r)
  toHtml (MTable rows) = mtable_ $ foldMap (mtr_ . foldMap (mtd_ . toHtml)) rows
  toHtml _ = mempty

  toHtmlRaw = toHtml -- TODO: provide unescaped HTML


type Parser = Parsec Void Text

sc :: Parser ()
sc = L.space space1 empty empty

lexeme :: Parser a -> Parser a
lexeme = L.lexeme sc

symbol :: Text -> Parser Text
symbol = L.symbol sc

brackets :: [(Text, Text)]
brackets = [ ("(", ")")
           , ("[", "]")
           , ("⟦", "⟧")
           , ("⦃", "⦄")
           , ("⟨", "⟩")
           ]

bracketedP :: Parser Exp
bracketedP = do
  (l, r) <- choice [ pure b <* string (fst b) | b <- brackets ]
  es <- sc *> rowP <* string r <* sc
  pure $ MRow ([MOp l] ++ es ++ [MOp r])

matP :: Parser Exp
matP = do
  "mat" <* sc
  (n, m) <-
    (,) <$ "(" <* sc
        <*> L.decimal <* sc <* "," <* sc
        <*> L.decimal <* sc <* ")" <* sc
  els <- replicateM n (replicateM m atomP)
  pure (MTable els)

-- atomic elements
atomP :: Parser Exp
atomP = bracketedP
    <|> ("{" *> sc *> expP <* "}" <* sc)
    <|> matP
    <|> try (MNum . pack . show <$> lexeme L.decimal)   -- integer (TODO: number literal)
    <|> try (MIdent <$> takeWhile1P Nothing isAlphaNum <* sc) -- identifier
    <|> try do
          c <- anySingle <* sc
          if c == '√' then do
            e <- atomP
            return (MRoot Nothing e)
          else if c `elem` ['+', '-', '∶', '⊢', '→'] then return (MOp (singleton c))
          else empty

fracP :: Parser Exp
fracP = do
  l <- atomP
  try (MFrac [] l <$> ("/" *> sc *> atomP)) <|> pure l

-- | parse a sequence of atomP elements
rowP :: Parser [Exp]
rowP = sc *> many fracP

seqP :: Parser Exp
seqP = do
  l <- rowP
  (do
    try $ some "─" <* sc
    r <- rowP
    pure (MUnderOver [class_ "seq"] (That (MRow l)) (MRow r))
    ) <|> pure (MRow l)

-- | Top-level exp parser
expP :: Parser Exp
expP = seqP -- <&> \case
  -- [ ] -> Empty
  -- [e] -> e
  -- xs  -> (MRow xs)

toExp :: Text -> Either String Exp
toExp src = 
  case runParser expP "math block" src of
    Left err -> Left (errorBundlePretty err)
    Right e  -> Right e

processInline :: MonadFail m => Inline -> m Inline
processInline (Math t e) =
  case toExp e of
    Left err -> fail err
    Right e  -> pure $
      RawInline (Format "html") $ toStrict $ renderText $
        math_ [ display_ (if t == DisplayMath then "block" else "inline") ] $ toHtml e
processInline i = pure i

processMath :: MonadFail m => Pandoc -> m Pandoc
processMath = walkM processInline
